### Routines to convert XDOC to PostScript

# created by Jim Davis April 27 1994
# Copyright (C) 1994 Xerox Incorporated

# Modified 6 May 94 to support superscripts

# Note that there's no reason to assume that the PostScript fonts will
# have the same widths as the characters in the document, so in
# general while the origin of each string will be correct the right
# edge will not be.  You see this clearly when you draw boxes around
# the text zones - the boxes are in the right places but the text extends
# beyond.  Also ScanWorX does not tell us which font was used, only
# the size, so e.g. Helvettica will be Roman.

# Maybe there should be an optional arg to specify whether this program
# should try to remain as faithful to the XDOC as possible (when used as
# debugging aid) or try to produce prettiest output possible (e.g.
# might extract figures from the TIFF and glue them in as images)

$version = "0.2";

# constants.
$page_width_inches = 8.5;
$page_height_inches = 11;
$ps_per_inch = 72;
$page_width_ps = $page_width_inches * $ps_per_inch;
$page_height_ps = $page_height_inches * $ps_per_inch;

# globals shared while processing files

# size of the XDOC page.  These values are just for illustrative purposes
$page_width = 2165;
$page_height = 2789;
$current_baseline = 0;
$font_name = "Times-Roman";
$font_size = 12;
$font_xheight = 0;
$font_fixed = 0;
$current_zone = 0;
$file_counter = 0;
# array of Boolean.
@zone_is_graphic = ();
@zone_is_text = ();


# Global input variables:
# $title
# $outline_text_zones
#  if 1, thin boxes will be drawn around all text zones
# $analyze_text_zones

# Caller can set $title if known.
sub XD_begin_document {
    local ($nfiles) = @_;
    print "%!PS-Adobe-2.0\n";
    print "%%Creator: xdoc-to-ps $version\n";
    if (defined ($title)) {
	print "%%Title: $title\n";}
#    print "%%Pages: $nfiles $nfiles\n";
    print "%%EndComments\n%%EndProlog\n";
}


sub XD_begin_of_file {
    local ($pass, $file, *fonts, *zones) = @_;
    if ($pass == 2) {
	$file_counter++;
	print "%%Page: $file_counter $file_counter\n";
	if ($analyze_text_zones) {&analyze_zones();}
	if ($outline_text_zones) {&outline_zones (*zones);}
    }}

sub XD_end_of_file {
    local ($pass, $file) = @_;
    if ($pass == 2) {
	print "showpage\n";}}

sub XD_beginpage {
    local ($pass, $left, $top, $width, $height) = @_;
    $page_width = $width;
    $page_height = $height;
    &initialize_zone_statistics;}

sub XD_beginline {
    local ($pass, $style, $zone, $left, $distance, $baseline) = @_;
    $current_baseline = $baseline;
    $current_zone = $zone;
    if ($pass == 2) {
	&Moveto (&xdoc_to_ps_coord ($left + $distance, $baseline));}}

# Ignore small whitespaces (less than width of one character) because
# the PostScript font metrics are not the same as those of the scanned
# page, and the small shifts will probably make things look worse not
# better.  On the other hand this   makes the output less faithful to
# the input.  Maybe it's not a good idea.
sub XD_whitespace {
    local ($pass, $left, $width) = @_;
    if ($pass == 2) {
	if ($width >= $font_xheight * .4) {
	    &Moveto (&xdoc_to_ps_coord ($left + $width, $current_baseline));}}}

sub XD_text {
    local ($pass, $string) = @_;
    if ($pass == 1) {
	if ($analyze_text_zones) {
	    &collect_zone_statistics ($current_zone, $string); }}
    else {
	&Show ($string);}}

$unrecognized_character = "?";

sub XD_unrecognized {
    if ($pass == 1) {
	&collect_zone_statistics_unrec ($current_zone);
    }
    else {
	&Show ($unrecognized_character);}}

sub XD_softhyphen {
    if ($pass == 2) {
	&Show ("-");}}

sub XD_endofline {
    if ($pass == 2) {
	print "\n";}}

# A font shift.
# I am told that findfont is fairly slow.  But efficiency comes last
# ScanWorX does not currently support the "serif" argument.
sub XD_font {
    local  ($pass, $face, $serif, $fixed, $point_size, 
	    $height_uc, $height_descender, $height_lc) = @_;

    $font_xheight = $height_lc;
    # fixed = F means fixed, V is variable.
    # face = B is Bold, R is Roman, I Italic

    $font_name = "Times-Roman";
    $font_size = $point_size;

    if ($fixed eq "F") {
 	$font_fixed = 1;
 	$font_name = "Courier";}
    else { 	
	$font_fixed = 2;
 	if ($face eq "R") {
	}
 	elsif ($face eq "B") {
	    $font_name = "Times-Bold"; 	}
	elsif ($face eq "I") {
	    $font_name = "Times-Italic";} }
    if ($pass == 2){
 	print "/$font_name findfont $point_size scalefont setfont\n";} }


# Output a small relative move, # and decrease the point size

$saved_point_size = 0;
$superscript_vertical_offset_factor = .5;
$superscript_point_size_factor = .75;

sub XD_superscript {
    local ($pass, $on) = @_;
    if ($pass == 2) {
 	if ($on) { 	
	    $saved_point_size = $font_size;
	    printf ("/$font_name findfont %d scalefont setfont\n",
		    int ($font_size * $superscript_point_size_factor));
	    printf ("0 %d rmoveto\n",
		    int ($font_size * $superscript_vertical_offset_factor));}
	else {
	    printf ("/$font_name findfont %d scalefont setfont\n",
		    $saved_point_size);
	    printf ("0 %d rmoveto\n",
		    int ($font_size * -$superscript_vertical_offset_factor));}
    }
}

# Add code.  implement with a small relative move
sub XD_subscript {
    local ($pass, $on) = @_;
}

sub XD_declare_graphic_zone {
    local ($pass, $zone, $compression, $left, $top, $width, $height) = @_;
    if ($pass == 2) {
	&Rect_path ($left, $top, $left+$width, $top+$height);
	print "closepath stroke\n";}}

## Zone analysis and display code

sub initialize_zone_statistics {
    @zone_is_graphic = ();
    @zone_is_text = ();
    @zc_total = ();
    @zc_alpha = ();
    @zc_unrec = ();
    @zc_other = ();}

sub collect_zone_statistics {
    local ($zone, $string) = @_;
    $zc_total[$zone] += length ($string);
    $zc_alpha[$zone] += ($string =~ tr/A-Za-z//d);
    $zc_other[$zone] += (length ($string));
    }

sub collect_zone_statistics_unrec {
    local ($zone) = @_;
    $zc_total[$zone]++;
    $zc_unrec[$zone]++;}


# global in @zc_total, @zc_unrec, $zc_unrec_cutoff

$zc_unrec_cutoff = .02;

# global out $zone_is_graphic
sub analyze_zones {
    print STDERR "Page $file_counter\n";
    for ($z = 0 ; $z <= $#zc_total; $z++) {
	if ($zc_total[$z] > 0) {
	    printf (STDERR "Zone %d total %d alpha %d unrec %d other %d",
		    $z, $zc_total[$z], $zc_alpha[$z], $zc_unrec[$z], $zc_other[$z]);
	    if (($zc_unrec[$z] / $zc_total[$z]) > $zc_unrec_cutoff) {
		printf (STDERR " graphic (%.2f)\n", ($zc_unrec[$z] / $zc_total[$z]));
		$zone_is_graphic[$z] = 1;}
	    elsif (($zc_other[$z] / $zc_total[$z]) < .08 &&
		   $zc_total[$z] > 32) {
		printf (STDERR " text (%.2f)\n", ($zc_other[$z] / $zc_total[$z]));
		$zone_is_text[$z] = 1;}  
	    else {
		print STDERR " uncertain\n";
	    }}}}

# global in @zone_is_graphic @zone_is_text
sub outline_zones {
    local (*zones) = @_;
    local ($zone) = 0;
    print "gsave 1 setlinewidth\n";
    foreach $z (@zones) {
	($left, $top, $right, $bottom) = (split (":", $z));
	if ($left && $top && $right && $bottom) {
	    &Rect_path ($left, $top, $right, $bottom);
	    if ($zone_is_graphic[$zone]) {
		print ".9 setgray closepath fill\n";}
	    elsif ($zone_is_text[$zone]) {
		print "0 setgray 1 setlinewidth closepath stroke\n";}
	    else {
		print ".9 setgray 1 setlinewidth closepath stroke\n";}
	}
	$zone++;
    }
    print "grestore\n";}


## Postscript routines


$PS_EM_DASH = "\320";
$PS_BULLET = "\267";

sub Show {
    local ($string) = @_;
    # quote any chars that have special meaning inside PS string
    $string =~ s/\\/\\\\/g;
    $string =~ s/\(/\\(/g;
    $string =~ s/\)/\\)/g;
    # XIS characters
    $string =~ s/$XIS_L_SINGLE_QUOTE/`/g;
    $string =~ s/$XIS_R_SINGLE_QUOTE/'/g;
    $string =~ s/$XIS_L_DOUBLE_QUOTE/"/g;
    $string =~ s/$XIS_R_DOUBLE_QUOTE/"/g;
    $string =~ s/$XIS_EM_DASH/$PS_EM_DASH/g;
    $string =~ s/$XIS_BULLET/$PS_BULLET/g;
    if (length ($string) > 0) {
	print "($string) show ";}}

sub Moveto {
    local ($px, $py) = @_;
    print "$px $py moveto ";}


# XDOC coordinates have origin in upper left, PS in lower left

sub xdoc_to_ps_coord {
    local ($x, $y) = @_;
    (int (($x / $page_width) * $page_width_ps),
     $page_height_ps - int (($y / $page_height) * $page_height_ps));}

sub Rect_path {
    local ($left, $top, $right, $bottom) = @_;
    local ($x, $y);
    &Moveto (&xdoc_to_ps_coord ($left, $top));
    ($x, $y) = &xdoc_to_ps_coord ($right, $top);
    print "$x $y lineto ";
    ($x, $y) = &xdoc_to_ps_coord ($right, $bottom);
    print "$x $y lineto ";
    ($x, $y) = &xdoc_to_ps_coord ($left, $bottom);
    print "$x $y lineto ";}


