### Routines to convert XDOC to HTML
# Copyright 1993, 1994 by Xerox Incorporated.
# created Jim Davis 20 Sept 93
# modified 5 Nov to fix bug in soft hyphenation, use <br> not <p>
# modified 27 Apr 94 for compatibility with new calling sequences.  No new features.


$font_fixed = 0;
$font_bold = 0;
$font_italic = 0;
$font_size = 12;
$font_xheight = 20;

# a word left over from previous line just before a soft hyphen
$leftover = "";

# store all characters from the current line here so that when the
# line ends we can inspect it to see whether it might be a heading
$current_line = "";

# Keep a count of the number of lines in this paragraph.  If it is
# less than min_para_lines, then don't make it a separate paragraph.
$current_line_cnt = 0;
$min_para_lines = 3;

# Keep track of the baseline so that we can (roughly) map the paragraphs
# back to the image.
$current_baseline = 0;

# And let's keep track of when we need to add the title block on (at the
# start of a new line).
$pending_title = 1;

# But if a font shift occurs within the line, then don't try.
# (Maybe later we'll try to handle this case.)  line_complete 0
# means that current_line does not have the complete line.
$line_complete = 1;

sub XD_beginline {
    local ($pass, $style, $zone, $left, $distance, $baseline) = @_;
    if ($pass == 2) {
	$current_line = "";
	$line_complete = 1;
	$current_line = $leftover;
	$leftover = "";
	if ($pending_title == 1){
	    $pending_title = 0;
	    print ".title\n";
	    print "$docid [$page_number;$current_baseline]\n";
	    print ".paragraph\n";}
	$current_baseline = $baseline;
    }}

# The OCR finds whitespaces within words, for example around hyphens.
# so we compare the width of the whitespace to the current x-height
# (the height of an x character).  For a fixed width font the space
# must be >= .75 the x-height.  For a variable width font the cutoff
# is much lower, because it's better to insert a spurious space than
# to throw one away.  The value of .35 comes because I saw at least
# one case where the ratio was .39  It may be that I will have to
# consider kerning.  This low value was an f followed by a t.
sub XD_whitespace {
    local ($pass, $left, $width) = @_;
    if ($pass == 2) {
	if ($width >= $font_xheight * ($font_fixed?.75:.35)) {
	    $current_line .= " ";}}}

sub XD_text {
    local ($pass, $_) = @_;
    if ($pass == 2) {
	$current_line .= $_;}}

$unrecognized_character = "?";

sub XD_unrecognized {
    local ($pass) = @_;
    if ($pass == 2) {
	$current_line .= $unrecognized_character;}}

sub XD_softhyphen {
    local ($pass) = @_;
    if ($pass == 2) {
	if ($current_line =~ /^(.*) ([^ ]+)$/) {
	    $current_line = $1;
	    $leftover = $2;
	} else {
	    $leftover = $current_line;}
	if ($leftover eq "") {
	    print STDERR "\"$current_line\"\n";
	    die "Could not find word before the soft hyphen in $file line $line_counter.\nStopped"; }}}


sub XD_endofline {
    local ($pass, $hard) = @_;
    if ($pass == 2) {
	if ($current_line ne "" &&
	    $line_complete) {
	    &AnalyzeLine($current_line,$hard);}
	else {
	    # incomplete line
	    print STDOUT $current_line;
	    if ($hard && $current_line_cnt >= $min_para_lines) {
		$current_line_cnt = 0;
		$pending_title = 1;
		print "\n";}
	    else {
	        if($current_line ne "") {
		    $current_line_cnt++;
		    print "\n";}}}}}

# missing features or bugs:  the last line on a page
# before the page number has a hard line ending.  it would
# be nice to recognize this and change it to soft. so as to
# merge in the next paragraph.

# Look for headers, etc in the line
sub AnalyzeLine {
    local ($line, $hard) = @_;
    local ($level,$number, $title) = &isHeaderLine($line);
    if ($level) {
	print "$title\n";}
    elsif ($line =~ /^[0-9]+$/ && $hard==1) {
	# suppress page numbers
    }
    else {
	print $line;
	if ($hard && $current_line_cnt >= $min_para_lines) {
	    $current_line_cnt = 0;
	    $pending_title = 1;
	    print "\n";}
	else {
	    $current_line_cnt++;
	    print "\n";}}}

sub isHeaderLine {
    local ($line) = @_;
    local ($level,$number,$title);
    if ($line =~ /^([\.0-9]+) +(.*)$/) {
	$number = $1;
	$title = $2;
	
	if ($title =~ /^[-A-Za-z,\(\) ]+$/) {
	    if ($number =~ /^[0-9][0-9]?$/) {
		$level = 2;}
	    elsif ($number =~ /^[0-9][0-9]?\.[0-9][0-9]?$/) {
		$level = 3;}
	}}
    ($level, $number, $title);}

# A font shift
sub XD_font {
    local  ($pass, $face, $serif, $fixed, $point_size, 
	    $height_uc, $height_descender, $height_lc) = @_;

    if ($pass == 2) {
	if ($current_line ne "") { 
	    # Got a font shift after we've already got some chars in the
	    # current line.  We don't have any way to store the font shift
	    # info in the line buffer, so we have to print it out right
	    # now.  But that means we need to print the line buffer too,
	    # to keep things in the right order, and that means that when
	    # we get to the end of the line we should not try to analyze
	    # the line because it will be only partial.
	    print STDOUT $current_line;
	    $current_line = "";
	    $line_complete = 0;}

	$font_size = $point_size;
	$font_xheight = $height_lc;

	# F means fixed, V is variable.
	# B is Bold, R is Roman, I Italic
	if ($fixed eq "F") {
	    if (! $font_fixed) {
		print "";}
	    $font_fixed = 1;}
	elsif ($fixed eq "V") {
	    if ($font_fixed) {
		print "";}
	    $font_fixed = 0;

	    if ($face eq "B") {
		if (! $font_bold) {
		    print "";
		    $font_bold = 1;}}

	    elsif ($face eq "I") {
		if (! $font_italic) {
		    print "";
		    $font_italic = 1;}}

	    elsif ($face eq "R") {
		if ($font_bold) {
		    print "";
		    $font_bold = 0;}
		if ($font_italic) {
		    print "";
		    $font_italic = 0;}
	    }}}}


